/*
    vbs - make mail attachments unexecutable

    Copyright 2000 Theo Nolte <nolte@post.rwth-aachen.de>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
*/

%{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <syslog.h>
#include <stdarg.h>
#include <sysexits.h>
#include <sys/utsname.h>
typedef unsigned char uchar;
enum {NBND=20};

struct mime_t {
  char*ctype;  /* content-type */
  char*ctenc;  /* content-transfer-encoding */
  char*cdesc;  /* content-description */
  char*cdisp;  /* content-disposition */
  char*cid;    /* content-id */
} mime;

struct patsub_t {
  char*pat;  /* extension */
  char*sub;  /* substitution */
} defsubs[]= {
  { ".bat", "~bat" },
  { ".com", "~com" },
  { ".exe", "~exe" },
  { ".reg", "~reg" },
  { ".scr", "~scr" },
  { ".shs", "~shs" },
  { ".vbe", "~vbe" },
  { ".vbs", "~vbs" },
  { 0,0 },
};

struct utsname utsbuf;

char*defbpat[]={0};

struct patsub_t*patsubs=defsubs;
FILE*fp;
char*progname,**bpat=defbpat;
int agentpid,dbg,m822,discard;
int DSTATE=1,DTRC=2;

int rdbndy(char*hdr) ; /* extract boundary */
char*getbndy(void) ;
int popbndy(void) ;
int nbndy(void) ;
void mangle(char*line,char*par) ;
void error(int exstat,int enr,char* msg,...) ;
void dp(int lvl,char*fmt,...) ; /* debug print */
char*getpar(char*hdr,char*name,int*off,int*len) ;
void domime(struct mime_t m) ;
char*strlow(char*s) ; /* return lower copy of s */
char*estrdup(char*s) ;
void manual(void);

%}

%x mhdr mbody thru

hdr .*(\n[ \t].*)*\n

%%

<thru>{
.|\n  {fputc(yytext[0],fp);}
}

<mhdr>{ /* mime header */
^content-type:{hdr}              { rdbndy(yytext); mime.ctype=estrdup(yytext);}
^content-disposition:{hdr}       { mime.cdisp=estrdup(yytext); }
^content-transfer-encoding:{hdr} { mime.ctenc=estrdup(yytext); }
^content-description:{hdr}       { mime.cdesc=estrdup(yytext); }
^content-id:{hdr}                { mime.cid=estrdup(yytext); }
^\r?\n   { domime(mime); memset(&mime,0,sizeof(mime));
           fwrite(yytext,yyleng,1,fp);
           if(m822) { /* recursive mime */
             m822=0;
             dp(DSTATE,"MHDR->MHDR");
           } else { /* body */
             dp(DSTATE,"MHDR->MBODY");
             BEGIN(mbody);
           }
         }
.|\n     { fputc(yytext[0],fp); }
}

<mbody>{ /* mime body */
.|\n  { if(!discard) fputc(yytext[0],fp);}
\r?\n--.*[^\r\n]  {
  char*b,*eob;
  int lb,off=yytext[0]=='\r';
  dp(DTRC,"\\n-- found: '%s'",yytext);
  if(!(b=getbndy())) { /* no mime left */
    fwrite(yytext,yyleng,1,fp);
    dp(DSTATE,"MBODY->THRU");
    BEGIN(thru);
  } else if(strncmp(yytext+3+off,b,lb=strlen(b))) { /* no match */;
    if(!discard) fwrite(yytext,yyleng,1,fp);
  } else if(!strcmp(eob=yytext+3+lb+off,"")) { /*boundary*/
    dp(DSTATE,"MBODY(%d)->MHDR(%d)",nbndy(),nbndy());
    fwrite(yytext,yyleng,1,fp);
    discard=0;
    BEGIN(mhdr);
  } else if(!strcmp(eob,"--")) { /* end boundary */
    fwrite(yytext,yyleng,1,fp);
    discard=0;
    if(popbndy()) { /* this mime done */
      dp(DSTATE,"MBODY(%d)->MBODY(%d)",nbndy()+1,nbndy());
    } else { /* all mime done */
      dp(DSTATE,"EOM");
      BEGIN(thru);
    }
  } else {
    if(!discard) fwrite(yytext,yyleng,1,fp);
    error(0,0,"bad boundary trailer");
  }
 }
}

%%

void error(int exstat,int enr,char* msg,...) { /* error to stderr or syslog */
 va_list ap; char s[512];
 int prio;
  prio=LOG_MAIL|(enr?LOG_ERR:LOG_WARNING);
  va_start(ap,msg);
  vsnprintf(s,sizeof(s),msg,ap);
  va_end(ap);
  if(enr) {
    char s2[512]; strncpy(s2,s,sizeof(s2)); s2[sizeof(s2)-1]='\0';
    snprintf(s,sizeof(s),"%s (%s)",s2,strerror(enr));
  }
  if(isatty(2)) fprintf(stderr,"%s: %s\n",progname,s);
    else syslog(prio,"%s",s);
  if(exstat) exit(exstat);
}

void dp(int lvl,char*fmt,...) { /* debug print */
 va_list ap;
 int nl=1;
 char*s;
  if(!(lvl&dbg)) return;
  va_start(ap,fmt);
  s=strdup(fmt);
  if(s[strlen(s)-1]=='-') nl=s[strlen(s)-1]=0;
  vfprintf(stderr,s,ap);
  if(nl) fprintf(stderr,"\n");
  va_end(ap);
  free(s);
}

void*emalloc(int n) {
 void*p=malloc(n);
  if(!p) error(EX_OSERR,errno,"out of mem while malloc'ing %d bytes",n);
  return(p);
}

char*estrdup(char*s) {
 char*p;
  if(!(p=strdup(s))) error(EX_OSERR,errno,"estrdup(%.20s): out of mem",s);
  return(p);
}

void*erealloc(char*p,int n) {
 void*q=realloc(p,n);
  if(!q) error(EX_OSERR,errno,"out of mem while realloc'ing %d bytes",n);
  return(q);
}

void outqstr(char*s) { /* output s as quoted string */
 char*p;
  dp(DTRC,"outqstr(%s)",s);
  fputc('"',fp);
  for(p=s;*p;p++)
    if(*p!='"'&&*p!='\\') fputc(*p,fp);
      else fprintf(fp,"\\%c",*p);
  fputc('"',fp);
}

char*qstrval(char*s,int*len) { /* read quoted string */
 char*p,*rv; int max,i;
  /*dp(DTRC,"qstrval(%s)-",s);*/
  rv=emalloc(max=1);
  if(*s!='"') { *rv='\0'; return(rv); }
  for(i=0,p=s+1;*p;p++) {
    if(i+1>=max) rv=erealloc(rv,max*=2);
    if(*p=='"') { p++; break; } else
    if(*p=='\\'&&*(p+1)) rv[i++]=*++p; else
    rv[i++]=*p;
  }
  rv[i]='\0'; *len=p-s;
/*  dp(DTRC,": '%s', len: %d",rv,*len);*/
  return(rv);
}

char*uqstrval(char*s,int*len) { /* read unquoted string */
 char*rv,*p,*tspecial="()<>@,;:\\\"/[]?=";
 int i,tsp[UCHAR_MAX+1];
  /*dp(DTRC,"uqstrval(%s)-",s);*/
  memset(tsp,'\0',sizeof(tsp));
  for(i=0;i<' ';i++) tsp[i]=1;
  for(p=tspecial;*p;p++) tsp[*(uchar*)p]=1;
  for(p=s;!tsp[*(uchar*)p];p++);
  rv=emalloc(p-s+1); memmove(rv,s,p-s); rv[p-s]='\0';
  *len=p-s;
  /* dp(DTRC,": '%s', len:%d",rv,*len);*/
  return(rv);
}

char*rvalue(char*s,int*len) { /* read value (quoted or unquoted) */
  return(*s=='"'?qstrval(s,len):uqstrval(s,len));
}

char*getpar(char*hdr,char*name,int*off,int*len) {
 char*p,*h,*pat;
 int o;
  if(!hdr) return NULL;
  h=strlow(hdr);
  if(!(pat=alloca(strlen(name)+2))) error(EX_OSERR,errno,"getpar: out of mem");
  sprintf(pat,"%s=",name);
  if((p=strstr(h,pat))) o=(*off=p-h)+strlen(pat);
  free(h);
  if(p) {
   char*rv;
    rv=rvalue(hdr+o,len);
    *len+=strlen(pat);
    while(*off>0&&isspace(hdr[*off-1])) { (*off)--; (*len)++; }
    if(*off>0&&hdr[*off-1]==';') { (*off)--; (*len)++; }
    return(rv);
  } else return(NULL);
  /* never reached */
}

char*strlow(char*s) { /* return lower copy of s */
 int i; char*p;
  p=emalloc(strlen(s)+1);
  for(i=0;s[i];i++) p[i]=tolower(s[i]);
  p[i]='\0';
  return(p);
}

int nbnd;
char*bndy[NBND];

int pushbndy(char*s) {
  if(nbnd>=NBND) error(0,0,"max mime recursion depth (%d) exceeded",NBND);
  bndy[nbnd++]=s;
  return(nbnd);
}

int popbndy(void) {
  if(nbnd>0) nbnd--;
  return(nbnd);
}

char*getbndy(void) {
  dp(DTRC,"getbndy()");
  return((nbnd>0)?bndy[nbnd-1]:NULL);
}

int nbndy(void) { return(nbnd); }

int rdbndy(char*hdr) { /* extract boundary */
 char*p,*ls,*b="boundary=";
 int l;
  dp(DTRC,"rdbndy(%s): entry",hdr);
  ls=strlow(hdr);
  m822=strstr(ls,"message/rfc822")!=NULL;
  if(!strstr(ls,"multipart")||!(p=strstr(ls,b))) { /* no boundary */
    free(ls);
    return(0);
  }
  pushbndy(rvalue(hdr+(p-ls+strlen(b)),&l));
  dp(DTRC,"rdbndy(%s): '%s'",hdr,getbndy());
  free(ls);
  return(1);
}

int subst(char**s) {
 struct patsub_t *ps;
 char*p,*q;
 int rv=0;
  if(!*s) return 0;
  q=strlow(*s);
  for(ps=patsubs;ps->pat;ps++) {
    while((p=strstr(q,ps->pat))) {
     char*n=emalloc(strlen(*s)+1+strlen(ps->sub)-strlen(ps->pat));
      rv=1;
      sprintf(n,"%.*s%s%s",p-q,*s,ps->sub,*s+(p-q+strlen(ps->pat)));
      free(*s); *s=n;
      free(q); q=strlow(*s);
    }
  }
  free(q);
  return(rv);
}

/* ensure parameter is unique */
void uniqpar(char*hdr,char*par) {
 char*v,*h;
 int o,l,o1,l1;
  if(!hdr) return;
  if(!(v=getpar(hdr,par,&o,&l))) return;
  free(v);
  h=hdr+o+l;
  while((v=getpar(h,par,&o1,&l1))) {
    free(v);
    while(o1>0&&isspace(h[o1-1])) { o1--; l1++; }
    if(o1>0&&h[o1-1]==';') { o1--; l1++; }
    memmove(h+o1,h+o1+l1,strlen(h+o1+l1)+1);
  }
}

void filter(struct mime_t m) {
 int c1,c2,o1,l1,o2,l2;
 char*v1,*v2;
  v1=getpar(m.ctype,"name",&o1,&l1);
  dp(DTRC,"filter: ct-val: '%s'",v1);
  v2=getpar(m.cdisp,"filename",&o2,&l2);
  dp(DTRC,"filter: cd-val: '%s'",v2);
  c1=subst(&v1);
  c2=subst(&v2);
  if(c1) {
    fprintf(fp,"%.*s; name=",o1,m.ctype);
    outqstr(v1);
    fprintf(fp,"%s",m.ctype+o1+l1);
  } else if(m.ctype) fprintf(fp,"%s",m.ctype);
  if(c2) {
    fprintf(fp,"%.*s; filename=",o2,m.cdisp);
    outqstr(v2);
    fprintf(fp,"%s",m.cdisp+o2+l2);
  } else if(m.cdisp) fprintf(fp,"%s",m.cdisp);
  if(m.ctenc) fprintf(fp,"%s",m.ctenc);
  if(m.cdesc) fprintf(fp,"%s",m.cdesc);
  if(m.cid)   fprintf(fp,"%s",m.cid);
  free(v1); free(v2);
  dp(DTRC,"filter: done");
}

void block(struct mime_t m,char*s) {
 char*eol;
  eol=m.ctype[strlen(m.ctype)-2]=='\r'?"\r\n":"\n";
  fprintf(fp,"Content-type: text/plain; charset=us-ascii%s",eol);
  fprintf(fp,"Content-disposition: inline%s%s",eol,eol);
  fprintf(fp,"Attachment \"%s\" blocked by vbs on host %s",s,utsbuf.nodename);
  discard=1;
}

int toblock(char*name) {
 char*s,**p; int rv=0;
  if(!name) return 0;
  s=strlow(name);
  for(p=bpat;*p;p++) if(strstr(s,*p)) { rv=1; break; }
  free(s);
  return(rv);
}

/* process mime headers */
void domime(struct mime_t m) {
 char*v1,*v2;
 int o1,l1,o2,l2;
  dp(DTRC,"domime: ctype: '%s'",m.ctype);
  dp(DTRC,"domime: ctenc: '%s'",m.ctenc);
  dp(DTRC,"domime: cdesc: '%s'",m.cdesc);
  dp(DTRC,"domime: cdisp: '%s'",m.cdisp);
  dp(DTRC,"domime: cid: '%s'",m.cid);
  uniqpar(m.ctype,"name");
  uniqpar(m.cdisp,"filename");
  v1=getpar(m.ctype,"name",&o1,&l1);
  dp(DTRC,"domime: ct-val: '%s'",v1);
  v2=getpar(m.cdisp,"filename",&o2,&l2);
  dp(DTRC,"domime: cd-val: '%s'",v2);
  if(toblock(v1)) block(m,v1);
  else if(toblock(v2)) block(m,v2);
  else filter(m);
  free(v1); free(v2);
}

void agent(char*proc,char**av) { /* fork delivery agent */
 int p[2];
  if(pipe(p)==-1) error(EX_OSERR,errno,"can't open pipe to %s",proc);
  if((fp=fdopen(p[1],"w"))==NULL) error(EX_OSERR,errno,"can't fdopen pipe");
  if((agentpid=fork())) { /* parent */
    close(p[0]);
  } else { /* child */
    close(p[1]);
    dup2(p[0],0);
    close(p[0]);
    if(execvp(proc,av)==-1) error(EX_OSERR,errno,"can't exec %s",proc);
  }
}

int wt(void) {
 int stat;
  waitpid(agentpid,&stat,0);
  return(WIFEXITED(stat)?WEXITSTATUS(stat):1);
}

void dpatsubs(void) { /* dump patsubs */
 struct patsub_t*ps;
  fprintf(stderr,"dpatsubs: ------ dump patsubs ------\n");
  for(ps=patsubs;ps->pat;ps++)
    fprintf(stderr,"dpatsubs: pat: '%s', sub: '%s'\n",ps->pat,ps->sub);
  fprintf(stderr,"dpatsubs: ------ dump done ------\n");
}

void dbpat(void) { /* dump bpat */
 char**p;
  fprintf(stderr,"dbpat: ------ dump bpat ------\n");
  for(p=bpat;*p;p++) fprintf(stderr,"dbpat: '%s'\n",*p);
  fprintf(stderr,"dbpat: ------ dump done ------\n");
}

void newblock(char*s) {
 static int len;
  if(!*s) return;
  if(bpat==defbpat) {
    bpat=emalloc((len=1)*sizeof(*bpat));
  }
  bpat=erealloc((void*)bpat,++len*sizeof(*bpat));
  bpat[len-2]=strlow(s);
  bpat[len-1]=NULL;
}

void newext(char*s) {
 static int len;
 char*p,*pat,*sub;
  dp(DTRC,"newext(%s)",s);
  if(!*s) return;
  if((p=strchr(s,'/'))) { /* pat/sub pair */
    pat=emalloc(p-s+1); strncpy(pat,s,p-s); pat[p-s]='\0';
    sub=emalloc(strlen(p+1)+1); strcpy(sub,p+1);
  } else { /* ordinary extension */
    pat=emalloc(strlen(s)+2);
    sprintf(pat,"%s%s",(*s=='.'?"":"."),s);
    sub=strlow(pat);
    sub[0]='~';
  }
  if(patsubs==defsubs) {
    len=1;
    patsubs=(struct patsub_t*)emalloc(len*sizeof(*patsubs));
  }
  patsubs=(struct patsub_t*)erealloc((void*)patsubs,++len*sizeof(patsubs[0]));
  patsubs[len-2].pat=strlow(pat);
  patsubs[len-2].sub=sub;
  patsubs[len-1].pat=NULL;
  free(pat);
}

char*uinfo[]={
"usage: vbs [vbs-options] agent [agent-options]",
"  vbs-options:",
"     -b blist   block any attachment whose filename contains one of the",
"                strings in blist, which is a colon-separated list of",
"                case-insensitive strings",
"     -d #       set debug level",
"     -e elist   rename attachments. elist is a colon-separated list of",
"                strings.  If an entry does not contain a '/', it is",
"                treated as an extension, else it is treated as a",
"                pat/sub pair, with the string pat to be substituted by",
"                sub.",
"     -h         print usage info",
"     -M         print manual",
"     -V         print version",
0,
};
void usage(void) {
 char**p;
  for(p=uinfo;*p;p++) printf("%s\n",*p);
  exit(0);
}

char**pargs(int ac,char**av) {
 int c;
 char*s,*t,*farg=0,*rarg=0;
  opterr=0;
  while((c=getopt(ac,av,"+b:d:e:f:hMr:V"))!=EOF) {
    switch(c) {
      case 'b': for(t=optarg;(s=strsep(&t,",: "));) newblock(s);
                break;
      case 'd': dbg=atoi(optarg); break;
      case 'e': for(t=optarg;(s=strsep(&t,",: "));) newext(s);
                break;
      case 'f': farg=optarg; break;
      case 'h': usage(); exit(0);
      case 'M': manual(); exit(0);
      case 'r': rarg=optarg; break;
      case 'V': printf("vbs 1.0.0\n"); exit(0);
      case '?': error(EX_USAGE,0,"bad option '%c'; try -h for help",optopt);
    }
  }
  if(optind>=ac) error(EX_USAGE,0,"no delivery agent specified");
  if(farg) { /* rearrange args for f */
    av[optind-2]=av[optind];
    av[optind-1]="-f";
    av[optind]=farg;
    optind-=2;
  }
  if(rarg) { /* rearrange args for r */
    av[optind-2]=av[optind];
    av[optind-1]="-r";
    av[optind]=rarg;
    optind-=2;
  }
  if(dbg&DTRC) { dpatsubs(); dbpat(); }
  return(av+optind);
}

int main(int ac, char**av) {
  progname=index(av[0],'/')?rindex(av[0],'/')+1:av[0];
  if(uname(&utsbuf)==-1) error(EX_OSERR,errno,"main: uname error");
  av=pargs(ac,av);
  agent(*av,av);
  BEGIN(mhdr);
  yylex();
  if(fclose(fp)) error(EX_OSERR,errno,"close error on stdout");
  return wt();
}
